// Multi Sweep, Duane B
// Using the servo library created by Michael Margolis 
// to control 12 Servos with one Arduino Uno

#include <Servo.h>
#include <RPLidar.h>

////CONFIG START

////DEBUG
#define DEBUG 1 //NULL --> off, 1 --> on Serial print, 2 --> es werden auch die Daten die Lidar ausgibt direkt mit ausgegeben
#define DEBUG_2 NULL //NULL --> off, 1 --> es werden auch die Daten die Lidar ausgibt direkt mit ausgegeben

#define SERVO_FOR_DATA 15 //Beispiel Servo 15 entspreicht 0 Grad Output für das 0 Grad servo, nur wenn DEBUG 1 

///GRUNDEINSTELLUN
#define SCAN_MIN_ANGLE 315 //min Scan angle
#define SCAN_MAX_ANGLE 45  // maximale Scan anlge
#define SCAN_RANGE 3 //delta Scan, hier immer 3 Rad fuer ein Servo
#define CONNECTED_SERVOS 30 //anzahl angeschlossner Servos
#define START_SERVO_PIN 23 //Servos beginnen ab diesen pin bis zu START_SERVO_PIN+CONNECTED_SERVOS (hier 40)
// macro just adds two - the first servo is attached to digital pin 2, this gives us upto 12 servos - digital 10 to 40
//erster Servon Pin, von da aus wird hochgezahlt z.B. 10 --> Servo 1 = Pin 10, Servo 2 = Pin 11...
#define RPLIDAR_MOTOR 3 // The PWM pin for control the speed of RPLIDAR's motor.
// This pin should connected with the RPLIDAR's MOTOCTRL signal 


//// FEINEINSTELLUNG
#define SERVO_UPDATE_RATE_LOOPS 200 //loops wie oft die servos geupdatet werden, dazwischen sammelt er Werte
#define MAX_RANGE 6000 //Maximale Scan range, alles was darueber ist, wird nicht genutzt
#define MIN_RANGE    0 //Minimale Scan range, alles was darunter ist, wird nicht genutzt
#define MIN_QUAL 5 //wenn die Qualitat niedriger als 20 ist, wird der Wert nicht genommen, und der alte behalten
#define DELTA_RANGE 100 //wenn das delta des neuen wertees groesser 100 ist, wird der alte Wert verwendet

////CONFIG ENDE


#define SERVO_TO_PIN(x) (x+START_SERVO_PIN)

Servo myServos[CONNECTED_SERVOS];
// You need to create an driver instance 
RPLidar lidar;

int array_values_angle[360];
int array_values_old_value[360];

int counter=0;

void setup()
{

  // attach the servos
  for(int nServo = 0;nServo < CONNECTED_SERVOS;nServo++)
  {
    myServos[nServo].attach(SERVO_TO_PIN(nServo));
  }

  // bind the RPLIDAR driver to the arduino hardware serial
  lidar.begin(Serial1);
  // set pin modes

  pinMode(RPLIDAR_MOTOR, OUTPUT);


  if(DEBUG or DEBUG_2){
    Serial.begin(9600);
    Serial.println("setup");   
  }

}



void loop()
{
  if (IS_OK(lidar.waitPoint())) {
    
    //Daten aus Lidar auslesen
    int distance = lidar.getCurrentPoint().distance; //distance value in mm unit
    int angle    = lidar.getCurrentPoint().angle; //anglue value in degree
    bool  startBit = lidar.getCurrentPoint().startBit; //whether this point is belong to a new scan
    byte  quality  = lidar.getCurrentPoint().quality; //quality of the current measurement

    if(DEBUG_2){
      ///Erster Winkel
      Serial.print("D:\t");
      Serial.print(distance);
      /// Distance vom ersten WINKEL
      Serial.print("\tA:\t");
      Serial.print(angle);
      /// Distance vom zweiten WINKEL
      Serial.print("\tSB\t");
      Serial.print(startBit);
      /// Distance vom dritten WINKEL
      Serial.print("\tQ:\t");
      Serial.println(quality);
    }
    
    
    // Pruefen der Qualtität und der Min/Max Range
    if(quality>=MIN_QUAL and distance<=MAX_RANGE and distance>=MIN_RANGE){
      //if(quality>=MIN_QUAL){
      ///alten Wert davor in einen Parallel Array speichern
      if(array_values_angle[angle] <= (array_values_old_value[angle]+DELTA_RANGE) and array_values_angle[angle] >= (array_values_old_value[angle]-DELTA_RANGE)){
        array_values_old_value[angle]=array_values_angle[angle];
      }      
      /// Werte des Radars in einen Array speichern bezogen auf 360 Werte 
      array_values_angle[angle]=distance;
    }


    // hochzahlen eines Counters, wenn dieser erreicht wird, werden die Servos wieder gestellt und der Counter auf 0 gesetzt
    if(counter >= SERVO_UPDATE_RATE_LOOPS){
      /// funktion Servo stellen
      move_servos();
      //counter auf 0 setzten
      counter=0;
    }
    //counter hochzählen
    counter++;


  } 
  else {
    analogWrite(RPLIDAR_MOTOR, 0); //stop the rplidar motor

    // try to detect RPLIDAR... 
    rplidar_response_device_info_t info;
    if (IS_OK(lidar.getDeviceInfo(info, 100))) {
      // detected...
      lidar.startScan();

      // start motor rotating at max allowed speed
      analogWrite(RPLIDAR_MOTOR, 255);
      delay(1000);
    }
  }

}




/////////////////////////////////////////////////////////7
//// FUNKTION SERVO STellen
void move_servos(){

  // Jedes der angeschlossen Servos durchlaufen
  for(int nServo = 0;nServo < CONNECTED_SERVOS;nServo++)
  {

    
    int angle_for_servo1=0;
    int angle_for_servo2=0;
    int angle_for_servo3=0;
    // fuer die erste hälfte der Servos werden die Werte zwischen 315 und 360 Grad genommen --> durch die Servo nummer wird der Gradbereich errechnet z.B. Servo 1 hat Gradbereich 3, 4, 5
    if(nServo<(CONNECTED_SERVOS/2)){                    
      angle_for_servo1=SCAN_MIN_ANGLE+0+nServo*SCAN_RANGE;
      angle_for_servo2=SCAN_MIN_ANGLE+1+nServo*SCAN_RANGE;
      angle_for_servo3=SCAN_MIN_ANGLE+2+nServo*SCAN_RANGE;
    }
    // fuer die zweite helfte der Servos werden die Werte von 0 bis 45 Grad genommen
    if(nServo>=(CONNECTED_SERVOS/2)){
      angle_for_servo1=(nServo-(CONNECTED_SERVOS/2))*SCAN_RANGE+0;
      angle_for_servo2=(nServo-(CONNECTED_SERVOS/2))*SCAN_RANGE+1;
      angle_for_servo3=(nServo-(CONNECTED_SERVOS/2))*SCAN_RANGE+2;
    }


    //Prufen, ob die der neue Wert innerhalb der Toleranz des maximalen deltas zum alten wert ist, wenn nicht einfach den alten wert nochmal verwenden
    if(array_values_angle[angle_for_servo1] >= (array_values_old_value[angle_for_servo1]+DELTA_RANGE) and array_values_angle[angle_for_servo1] <= (array_values_old_value[angle_for_servo1]-DELTA_RANGE)){
      array_values_angle[angle_for_servo1]=array_values_old_value[angle_for_servo1];
    } 

    //Prufen, ob die der neue Wert innerhalb der Toleranz des maximalen deltas zum alten wert ist, wenn nicht einfach den alten wert nochmal verwenden
    if(array_values_angle[angle_for_servo2] >= (array_values_old_value[angle_for_servo2]+DELTA_RANGE) and array_values_angle[angle_for_servo2] <= (array_values_old_value[angle_for_servo2]-DELTA_RANGE)){
      array_values_angle[angle_for_servo2]=array_values_old_value[angle_for_servo2];
    } 

    //Prufen, ob die der neue Wert innerhalb der Toleranz des maximalen deltas zum alten wert ist, wenn nicht einfach den alten wert nochmal verwenden
    if(array_values_angle[angle_for_servo3] >= (array_values_old_value[angle_for_servo3]+DELTA_RANGE) and array_values_angle[angle_for_servo3] <= (array_values_old_value[angle_for_servo3]-DELTA_RANGE)){
      array_values_angle[angle_for_servo3]=array_values_old_value[angle_for_servo3];
    }     

    //Mittelwert aus den drei distancen erstellen
    int mittelwert_distance=(array_values_angle[angle_for_servo1]+array_values_angle[angle_for_servo2]+array_values_angle[angle_for_servo3])/SCAN_RANGE;

    //Übertragen der Distance auf die Servo Range also 0 bis 180 Grad, wobei die MIN und MAX Range berücksichtigt wird, alles was darüber ist oder darunter wird wie MIN/MAx gewertet
    float servo_move_angle = map(mittelwert_distance, MIN_RANGE, MAX_RANGE, 0, 180);

    //Stellen des Servos
    myServos[nServo].write(servo_move_angle);

    //OUTPUT DATEN zur INFORMATION eines SERVOS um zu kontrollieren
    if(nServo==SERVO_FOR_DATA and DEBUG){
      ///Erster Winkel
      Serial.print("W:\t");
      Serial.print(angle_for_servo1);
      /// Distance vom ersten WINKEL
      Serial.print("\tD1:\t");
      Serial.print(array_values_angle[angle_for_servo1]);
      /// Distance vom zweiten WINKEL
      Serial.print("\tD2\t");
      Serial.print(array_values_angle[angle_for_servo2]);
      /// Distance vom dritten WINKEL
      Serial.print("\tD3:\t");
      Serial.print(array_values_angle[angle_for_servo3]);
      /// MITTELWERT der Drei Distancen
      Serial.print("\tMW:\t");
      Serial.print(mittelwert_distance);
      //SERVONUMMER
      Serial.print("\tServo:\t");
      Serial.print(nServo);
      ///STELLWINKEL des SERVOS
      Serial.print("\tAngle:\t");
      Serial.println(servo_move_angle);
    }
  }



}

